# distutils: language = c++
# cython: embedsignature=True

from libcpp.string cimport string
from libcpp.vector cimport vector
from cython.operator cimport dereference as deref
import numpy as np
cimport numpy as np
from cpython cimport PyObject, Py_INCREF

cimport cpp
from cpp cimport GTAR as GTAR_

np.import_array()

cdef class OpenMode:
    """
    Enum for ways in which an archive file can be opened

       .. data:: Read
       .. data:: Write
       .. data:: Append
    """
    Read = cpp.Read
    Write = cpp.Write
    Append = cpp.Append

cdef class CompressMode:
    """
    Enum for ways in which files within an archive can be compressed

       .. data:: NoCompress
       .. data:: FastCompress
       .. data:: MediumCompress
       .. data:: SlowCompress"""
    NoCompress = cpp.NoCompress
    FastCompress = cpp.FastCompress
    MediumCompress = cpp.MediumCompress
    SlowCompress = cpp.SlowCompress

cdef class Behavior:
    """
    Enum for how properties can behave over time

       .. data:: Constant
       .. data:: Discrete
       .. data:: Continuous"""
    Constant = cpp.Constant
    Discrete = cpp.Discrete
    Continuous = cpp.Continuous

cdef class Format:
    """
    Formats in which binary properties can be stored

       .. data:: Float32
       .. data:: Float64
       .. data:: Int32
       .. data:: Int64
       .. data:: UInt8
       .. data:: UInt32
       .. data:: UInt64"""
    Float32 = cpp.Float32
    Float64 = cpp.Float64
    Int32 = cpp.Int32
    Int64 = cpp.Int64
    UInt8 = cpp.UInt8
    UInt32 = cpp.UInt32
    UInt64 = cpp.UInt64

cdef class Resolution:
    """
    Resolution at which properties can be recorded

       .. data:: Text
       .. data:: Uniform
       .. data:: Individual"""
    Text = cpp.Text
    Uniform = cpp.Uniform
    Individual = cpp.Individual

cdef string py3str(arg):
    return <string> arg.encode('utf8')

cdef unpy3str(arg):
    return arg.decode('utf8')

widths = {'position': 3,
          'velocity': 3,
          'orientation': 4,
          'image': 3,
          'angular_momentum': 3,
          'angular_momentum_quat': 4,
          'moment_inertia': 3,
          'virial': 6,
          'center_of_mass': 3}

cdef class SharedArray:
    """Wrapper for the c++ SharedArray<char> class"""
    cdef cpp.SharedArray[char] *thisptr

    def __cinit__(self):
        """Initialize an empty SharedArray object"""
        self.thisptr = new cpp.SharedArray[char]()

    def __dealloc__(self):
        """Deallocate the reference to the currently-held SharedArray
        object"""
        del self.thisptr

    cdef copy(self, const cpp.SharedArray[char] &other):
        """Make this SharedArray point to the given target"""
        self.thisptr.copy(other)

    def __len__(self):
        """Return the number of elements (characters) in this
        SharedArray"""
        return self.thisptr.size()

    def __getitem__(self, index):
        """Access to individual elements in the SharedArray"""
        if(index < self.thisptr.size()):
            return self.thisptr.get()[index]
        else:
            raise IndexError('Index out of range')

    def __array__(self):
        """Interface to create a numpy array from this object"""
        cdef np.npy_intp shape[1]
        shape[0] = <np.npy_intp> self.thisptr.size()
        result = np.PyArray_SimpleNewFromData(1, shape,
                                            np.NPY_UINT8, self.thisptr.get())
        self._setBase(result)

        return result

    def __bytes__(self):
        return <bytes> self.thisptr.get()[:self.thisptr.size()]

    def __str__(self):
        return unpy3str(self.__bytes__())

    def _setBase(self, arr):
        """Sets the base of arr to be this object and increases the
        reference count"""
        cpp.PyArray_SetBaseObject(arr, self)
        Py_INCREF(self)

    def _arrayRecord(self, rec):
        """Create a numpy array from this object given some metadata in a Record
        object. Uses the widths dictionary in the gtar namespace to
        reshape the array into an Nxwidths[prop] array if the property's
        name is present there."""
        buf = np.asarray(self)

        formats = {cpp.Float32: '<f4',
                   cpp.Float64: '<f8',
                   cpp.Int32: '<i4',
                   cpp.Int64: '<i8',
                   cpp.UInt32: '<u4',
                   cpp.UInt64: '<u8',
                   cpp.UInt8: 'c'}

        result = np.frombuffer(buf, dtype=formats[rec.getFormat()])

        name = rec.getName()
        if name in widths:
            result = result.reshape((-1, widths[name]))

        return result

cdef class Record:
    """Python wrapper for the c++ Record class. Provides basic access to
    Record methods. Initializes in different ways depending on the
    number of given parameters.

    - No arguments: default constructor
    - 1 argument: Parse the given path
    - 6 arguments: Fill each field of the Record object (*group*, *name*, *index*, *behavior*, *format*, *resolution*)
    """

    cdef cpp.Record *thisptr

    def __cinit__(self, *args):
        if len(args) == 0:
            self.thisptr = new cpp.Record()
        elif len(args) == 1:
            self.thisptr = new cpp.Record(py3str(args[0]))
        elif len(args) == 6:
            self.thisptr = new cpp.Record(py3str(args[0]), py3str(args[1]),
                                          py3str(args[2]),
                                          args[3], args[4], args[5])
        else:
            raise TypeError('Incorrect number of arguments to Record()')

    def __dealloc__(self):
        """Clean up the record allocated for this object"""
        del self.thisptr

    def __richcmp__(self, other, int op):
        self_vals = [self.getGroup(), self.getName(), self.getIndex(),
                     self.getBehavior(), self.getFormat(), self.getResolution()]
        other_vals = [self.getGroup(), self.getName(), self.getIndex(),
                     self.getBehavior(), self.getFormat(), self.getResolution()]
        if op == 0:
            return self_vals < other_vals
        elif op == 1:
            return self_vals <= other_vals
        elif op == 2:
            return self_vals == other_vals
        elif op == 3:
            return self_vals != other_vals
        elif op == 4:
            return self_vals > other_vals
        elif op == 5:
            return self_vals >= other_vals
        else:
            raise TypeError('Unknown operator {} in __richcmp__'.format(op))

    def __hash__(self):
        return hash((self.__class__, repr(self)))

    def __reduce__(self):
        return (self.__class__, (self.getPath(),))

    def __repr__(self):
        """Returns a string representation of this object"""
        return 'Record("{}")'.format(self.getPath())

    cdef copy(self, const cpp.Record &other):
        """Set this object to be a copy of the given object"""
        self.thisptr.copy(other)

    def nullifyIndex(self):
        """Nullify the index field of this object"""
        return unpy3str(self.thisptr.nullifyIndex())

    def getBehavior(self):
        """Returns the `behavior` field of this object"""
        return self.thisptr.getBehavior()

    def getGroup(self):
        """Returns the `group` field of this object"""
        return unpy3str(self.thisptr.getGroup())

    def getName(self):
        """Returns the `name` field of this object"""
        return unpy3str(self.thisptr.getName())

    def getFormat(self):
        """Returns the `format` field of this object"""
        return self.thisptr.getFormat()

    def getPath(self):
        """Generates the path of the file inside the archive for this object"""
        return unpy3str(self.thisptr.getPath())

    def getResolution(self):
        """Returns the `resolution` field for this object"""
        return self.thisptr.getResolution()

    def getIndex(self):
        """Returns the `index` field for this object"""
        return unpy3str(self.thisptr.getIndex())

    def setIndex(self, index):
        """Sets the `index` field of this object"""
        self.thisptr.setIndex(py3str(index))

cdef class BulkWriter:
    """Class for efficiently writing multiple records at a time. Works
    as a context manager.

    :param arch: :py:class:`gtar.GTAR` archive object to write within

    Example::

        with gtar.GTAR('traj.sqlite', 'w') as traj, traj.getBulkWriter() as writer:
            writer.writeStr('notes.txt', 'example text')

    """
    cdef GTAR_.BulkWriter *thisptr

    def __cinit__(self, GTAR arch):
        self.thisptr = new GTAR_.BulkWriter(deref(arch.thisptr))

    cdef _dealloc(self):
        del self.thisptr
        self.thisptr = <GTAR_.BulkWriter*> 0

    def __dealloc__(self):
        self._dealloc()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._dealloc()

    def writeBytes(self, path, contents, mode=cpp.FastCompress):
        """Write the given contents to the location within the
        archive, using the given compression mode.

        :param path: Path within the archive to write
        :param contents: Bytestring to write
        :param mode: Optional compression mode (defaults to fast compression)
        """
        self.thisptr.writeString(py3str(path), contents, mode)

    def writeStr(self, path, contents, mode=cpp.FastCompress):
        """Write the given string to the given path, optionally
        compressing with the given mode.

        :param path: Path within the archive to write
        :param contents: String to write
        :param mode: Optional compression mode (defaults to fast compression)

        Example::

            writer.writeStr('params.json', json.dumps(params))
        """
        self.writeBytes(path, contents.encode('utf-8'), mode)

    def writePath(self, path, contents, mode=cpp.FastCompress):
        """Writes the given contents to the given path, converting as
        necessary.

        :param path: Path within the archive to write
        :param contents: Object which can be converted into array or string form, based on the given path
        :param mode: Optional compression mode (defaults to fast compression)
        """
        rec = Record(path)
        self.writeRecord(rec, contents, mode)

    def writeArray(self, path, arr, mode=cpp.FastCompress, dtype=None):
        """Write the given numpy array to the location within the
        archive, using the given compression mode. This serializes the
        data into the given binary data type or the same binary format
        that the numpy array is using.

        :param path: Path within the archive to write
        :param arr: Array-like object
        :param mode: Optional compression mode (defaults to fast compression)
        :param dtype: Optional numpy dtype to force conversion to

        Example::

            writer.writeArray('diameter.f32.ind', numpy.ones((N,)))
        """
        arr = np.ascontiguousarray(np.asarray(arr).flat, dtype=dtype)
        cdef np.ndarray[char, ndim=1, mode="c"] carr = np.frombuffer(arr, dtype=np.uint8)
        if carr.nbytes:
            self.thisptr.writePtr(py3str(path), &carr[0], carr.nbytes, mode)
        else:
            self.thisptr.writePtr(py3str(path), <void*> 0, carr.nbytes, mode)

    def writeRecord(self, Record rec, contents, mode=cpp.FastCompress):
        """Writes the given contents to the path specified by the given record.

        :param rec: :py:class:`gtar.Record` object specifying the record
        :param contents: [byte]string or array-like object to write
        :param mode: Optional compression mode (defaults to fast compression)
        """
        dtypes = {cpp.Float32: np.float32,
                  cpp.Float64: np.float64,
                  cpp.Int32: np.int32,
                  cpp.Int64: np.int64,
                  cpp.UInt8: np.uint8,
                  cpp.UInt32: np.uint32,
                  cpp.UInt64: np.uint64}

        if rec.getResolution() == cpp.Text:
            if type(contents) == str:
                self.writeStr(rec.getPath(), contents, mode)
            elif type(contents) == bytes:
                self.writeBytes(rec.getPath(), contents, mode)
            else:
                raise RuntimeError('Not sure how to serialize the given argument!')
        else:
            self.writeArray(rec.getPath(), contents, mode,
                            dtype=dtypes[rec.getFormat()])

cdef class GTAR:
    """Python wrapper for the :cpp:class:`GTAR` c++ class. Provides
    basic access to its methods and simple methods to read and write
    files within archives.

    The backend is automatically selected based on the suffix of the
    given path: if the name ends in '.tar', a tar-format archive will
    be created, if it ends in '.sqlite' a sqlite-format archive will
    be created, if it ends in '/' a directory structure (filesystem)
    "archive" will be created, otherwise a zip-format archive will be
    created.

    The open mode controls how the file will be opened.

    - read: The file will be opened in read-only mode
    - write: A new file will be opened for writing, potentially overwriting an existing file of the same name
    - append: A file will be opened for writing, adding to the end of a file if it already exists with the same name

    :param path: Path to the file to open
    :param mode: Open mode: one of 'r', 'w', 'a'
    """
    cdef cpp.GTAR *thisptr
    cdef _path
    cdef _mode

    openModes = {'r': cpp.Read,
                 'w': cpp.Write,
                 'a': cpp.Append}

    def __cinit__(self, path, mode):
        """Initialize a `GTAR` object given an archive path and open mode"""
        self._path = path
        self._mode = mode
        try:
            self.thisptr = new cpp.GTAR(py3str(path), self.openModes[self._mode])
        except KeyError:
            raise RuntimeError('Unknown open mode: {}'.format(self._mode))
        except RuntimeError as e:
            raise RuntimeError('{} for file {}'.format(', '.join(e.args), path))

    def __dealloc__(self):
        """Destroy the held `GTAR` object"""
        del self.thisptr

    def __enter__(self):
        """Enter a context with this object"""
        return self

    def __exit__(self, type, value, traceback):
        """Exit a context with this object"""
        self.close()

    def __reduce__(self):
        return (self.__class__, (self._path, self._mode))

    def _sortFrameKey(self, v):
        """Key function for sorting frame indices"""
        return (len(v), v)

    def close(self):
        """Close the file this object is writing to. It is safe to
        close a file multiple times, but impossible to read from or
        write to it after closing."""
        self.thisptr.close()

    def readBytes(self, path):
        """Read the contents of the given location within the archive,
        or return ``None`` if not found

        :param path: Path within the archive to write
        """
        result = SharedArray()
        result.copy(self.thisptr.readBytes(py3str(path)))

        return (bytes(result) if len(result) else None)

    def writeBytes(self, path, contents, mode=cpp.FastCompress):
        """Write the given contents to the location within the
        archive, using the given compression mode.

        :param path: Path within the archive to write
        :param contents: Bytestring to write
        :param mode: Optional compression mode (defaults to fast compression)
        """
        self.thisptr.writeString(py3str(path), contents, mode)

    def readStr(self, path):
        """Read the contents of the given path as a string or return
        ``None`` if not found.

        :param path: Path within the archive to write
        """
        result = self.readBytes(path)
        if result is not None:
            return result.decode('utf8')
        else:
            return result

    def writeStr(self, path, contents, mode=cpp.FastCompress):
        """Write the given string to the given path, optionally
        compressing with the given mode.

        :param path: Path within the archive to write
        :param contents: String to write
        :param mode: Optional compression mode (defaults to fast compression)

        Example::

            gtar.writeStr('params.json', json.dumps(params))
        """
        self.writeBytes(path, contents.encode('utf-8'), mode)

    def readPath(self, path):
        """Reads the contents of a record at the given path. Returns ``None`` if
        not found. If an array is found and the property is present in
        ``gtar.widths``, reshape into an Nxwidths[prop] array.

        :param path: Path within the archive to write
        """
        rec = Record(path)
        return self.getRecord(rec, rec.getIndex())

    def writePath(self, path, contents, mode=cpp.FastCompress):
        """Writes the given contents to the given path, converting as
        necessary.

        :param path: Path within the archive to write
        :param contents: Object which can be converted into array or string form, based on the given path
        :param mode: Optional compression mode (defaults to fast compression)
        """
        rec = Record(path)
        self.writeRecord(rec, contents, mode)

    def writeArray(self, path, arr, mode=cpp.FastCompress, dtype=None):
        """Write the given numpy array to the location within the
        archive, using the given compression mode. This serializes the
        data into the given binary data type or the same binary format
        that the numpy array is using.

        :param path: Path within the archive to write
        :param arr: Array-like object
        :param mode: Optional compression mode (defaults to fast compression)
        :param dtype: Optional numpy dtype to force conversion to

        Example::

            gtar.writeArray('diameter.f32.ind', numpy.ones((N,)))
        """
        arr = np.ascontiguousarray(np.asarray(arr).flat, dtype=dtype)
        cdef np.ndarray[char, ndim=1, mode="c"] carr = np.frombuffer(arr, dtype=np.uint8)
        if carr.nbytes:
            self.thisptr.writePtr(py3str(path), &carr[0], carr.nbytes, mode)
        else:
            self.thisptr.writePtr(py3str(path), <void*> 0, carr.nbytes, mode)

    def getBulkWriter(self):
        """Get a :py:class:`gtar.BulkWriter` context object. These allow for more
        efficient writes when writing many records at once."""
        return BulkWriter(self)

    def getRecord(self, Record query, index=""):
        """Returns the contents of the given base record and index.

        :param query: Prototypical :py:class:`gtar.Record` object describing the record to fetch
        :param index: Index used to fetch the record (defaults to index embedded in :code:`query`)
        :type query: :py:class:`gtar.Record`
        :type index: string

        .. note::

           If an index is passed into this function, it takes precedence over the index embedded in the given record.

        """
        rec = Record()
        rec.copy(deref(query.thisptr))

        if index != "":
            rec.setIndex(index)

        cdef cpp.SharedArray[char] inter
        inter = self.thisptr.readBytes(rec.thisptr.getPath())
        result = SharedArray()
        result.copy(inter)

        if rec.thisptr.getResolution() != cpp.Text:
            return result._arrayRecord(rec)
        else:
            try:
                return str(result)
            except UnicodeDecodeError:
                return bytes(result)

    def writeRecord(self, Record rec, contents, mode=cpp.FastCompress):
        """Writes the given contents to the path specified by the given record.

        :param rec: :py:class:`gtar.Record` object specifying the record
        :param contents: [byte]string or array-like object to write
        :param mode: Optional compression mode (defaults to fast compression)
        """
        dtypes = {cpp.Float32: np.float32,
                  cpp.Float64: np.float64,
                  cpp.Int32: np.int32,
                  cpp.Int64: np.int64,
                  cpp.UInt8: np.uint8,
                  cpp.UInt32: np.uint32,
                  cpp.UInt64: np.uint64}

        if rec.getResolution() == cpp.Text:
            if type(contents) == str:
                self.writeStr(rec.getPath(), contents, mode)
            elif type(contents) == bytes:
                self.writeBytes(rec.getPath(), contents, mode)
            else:
                raise RuntimeError('Not sure how to serialize the given argument!')
        else:
            self.writeArray(rec.getPath(), contents, mode,
                            dtype=dtypes[rec.getFormat()])

    def getRecordTypes(self, group=None, group_prefix=None):
        """Returns a python list of all the record types (without index
        information) available in this archive. Optionally filters
        results down to records found with a particular group name, if
        requested.

        :param group: Exact group name to select (default: do not filter by group); overrules `group_prefix`
        :param group_prefix: Prefix of group name to select (default: do not filter by group)

        """
        result = []
        types = self.thisptr.getRecordTypes()
        if group is not None:
            for rec in types:
                if unpy3str(rec.getGroup()) == group:
                    copy = Record()
                    copy.copy(rec)
                    result.append(copy)
        elif group_prefix is not None:
            for rec in types:
                if unpy3str(rec.getGroup()).startswith(group_prefix):
                    copy = Record()
                    copy.copy(rec)
                    result.append(copy)
        else:
            for rec in types:
                copy = Record()
                copy.copy(rec)
                result.append(copy)
        return result

    def queryFrames(self, Record target):
        """Returns a python list of all indices associated with a
        given record available in this archive

        :param target: Prototypical :py:class:`gtar.Record` object (the index of which is unused)
        """
        result = []
        frames = self.thisptr.queryFrames(deref(target.thisptr))
        for f in frames:
            result.append(unpy3str(f))
        return result

    def framesWithRecordsNamed(self, names, group=None, group_prefix=None):
        """Returns ``([record(val) for val in names], [frames])`` given a
        set of record names names. If only given a single name,
        returns ``(record, [frames])``.

        :param names: Iterable object yielding a set of property names
        :param group: Exact group name to select (default: do not filter by group); overrules `group_prefix`
        :param group_prefix: Prefix of group name to select (default: do not filter by group)
        """
        allRecords = dict((rec.getName(), rec) for rec in
            self.getRecordTypes(group=group, group_prefix=group_prefix))

        frames = None
        records = []

        if type(names) == type(''):
            names = [names]

        for n in names:
            try:
                rec = allRecords[n]
                records.append(rec)
            except KeyError:
                return (None, [])

            f = self.queryFrames(rec)

            if frames is not None:
                frames.intersection_update(f)
            else:
                frames = set(f)

        if len(names) > 1:
            return (records, sorted(frames, key=self._sortFrameKey))
        else:
            return (records[0], sorted(frames, key=self._sortFrameKey))

    def recordsNamed(self, names, group=None, group_prefix=None):
        """Returns ``(frame, [val[frame] for val in names])`` for each frame which
        contains records matching each of the given names. If only given
        a single name, returns ``(frame, val[frame])`` for each found
        frame. If a property is present in ``gtar.widths``, returns it
        as an Nxwidths[prop] array.

        :param names: Iterable object yielding a set of property names
        :param group: Exact group name to select (default: do not filter by group); overrules `group_prefix`
        :param group_prefix: Prefix of group name to select (default: do not filter by group)

        Example::

            g = gtar.GTAR('dump.zip', 'r')

            # grab single property
            for (_, vel) in g.recordsNamed('velocity'):
                pass

            # grab multiple properties
            for (idx, (pos, quat)) in g.recordsNamed(['position', 'orientation']):
                pass
        """
        allRecords = dict((rec.getName(), rec) for rec in
            self.getRecordTypes(group=group, group_prefix=group_prefix))

        frames = None

        if type(names) == type(''):
            names = [names]

        for n in names:
            try:
                rec = allRecords[n]
            except KeyError:
                return []
                # raise KeyError('Can\'t find a record named {}'.format(n))

            f = self.queryFrames(rec)

            if frames is not None:
                frames.intersection_update(f)
            else:
                frames = set(f)

        if len(names) > 1:
            for frame in sorted(frames, key=self._sortFrameKey):
                yield frame, tuple(self.getRecord(allRecords[n], frame) for n in names)
        else:
            for frame in sorted(frames, key=self._sortFrameKey):
                yield frame, self.getRecord(allRecords[names[0]], frame)

    def staticRecordNamed(self, name, group=None, group_prefix=None):
        """Returns a static record with the given name. If the property is
        found in ``gtar.widths``, returns it as an Nxwidths[prop]
        array. Optionally restricts the search to records with the
        given group name or group name prefix.

        :param name: Name of the property to find
        :param group: Exact group name to select (default: do not filter by group); overrules `group_prefix`
        :param group_prefix: Prefix of group name to select (default: do not filter by group)

        """
        try:
            rec = [rec for rec in
                   self.getRecordTypes(group=group, group_prefix=group_prefix)
                   if rec.getName() == name][0]
            return self.getRecord(rec)
        except IndexError:
            raise KeyError('Can\'t find a static record named {}'.format(name))

def isZip64(filename):
    """Internal helper function. Returns ``True`` if a file located at the
    given path is in zip64 format."""
    return cpp.isZip64(py3str(filename))




"""
Fix a getar-formatted zip file.

::

   usage: python -m gtar.fix [-h] [-o OUTPUT] input

   Command-line zip archive fixer

   positional arguments:
     input                 Input zip file to read

   optional arguments:
     -h, --help            show this help message and exit
     -o OUTPUT, --output OUTPUT
   Output location for fixed zip archive
"""

import argparse
import os
import re
import subprocess
import sys

from . import GTAR
from . import copy
from . import isZip64

parser = argparse.ArgumentParser(
    description='Command-line zip archive fixer')
parser.add_argument('input',
                    help='Input zip file to read')
parser.add_argument('-o', '--output', default='fixed.zip',
                    help='Output location for fixed zip archive')

def rebuildZip(input, tempName):
    cmdLine = ['zip', '-FF', input, '--out', tempName]

    proc = subprocess.Popen(cmdLine, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    (zipLog, _) = proc.communicate('y\n'.encode('UTF-8'))
    zipLog = zipLog.decode('utf8').split('\n')
    zipLog.pop()

    if proc.returncode not in [0, 3]:
        raise RuntimeError('Error {} running zip command'.format(proc.returncode))

    return zipLog

def deleteBadFrames(tempName, zipLog):
    zipLog = list(zipLog)
    deleteLines = []
    # Bad zip records from miniz always have this size [citation needed]
    if zipLog[-1].strip().endswith('(4294967295 bytes)'):
        foundFrame = re.search('frames/([0-9]+)/', zipLog[-1])

        if foundFrame:
            frame = foundFrame.group(0)
            while frame in zipLog[-1]:
                deleteLines.append(zipLog.pop())
        else:
            deleteLines.append(zipLog.pop())

    toDelete = []
    pattern = re.compile('^\\s*copying: (\\S*)\\s+\\(\\d+ bytes\\)$', flags=re.UNICODE)
    for line in deleteLines:
        res = pattern.match(line)
        if res:
            toDelete.append(res.group(1))

    if toDelete:
        print('Removing the following files: {}'.format(', '.join(toDelete)))
        cmdLine = ['zip', '-d', tempName] + toDelete
        subprocess.check_output(cmdLine)

    return toDelete

def main(input, output):
    """Fix a getar-formatted zip file.

    :param input: Input filename
    :param output: Output filename (can be the same as input)

    """

    if input.endswith('.tar') or input.endswith('.sqlite'):
        with gtar.GTAR(input, 'a'):
            pass
        return

    needsFix = True

    # if it ain't broke, don't fix it
    try:
        with GTAR(input, 'r'):
            needsFix = False
            if os.path.samefile(input, output) and isZip64(output):
                return 0
    except Exception:
        pass

    nameHalves = os.path.splitext(output)
    tempName = output

    if not os.path.exists(input):
        raise RuntimeError('Can\'t find input file {}'.format(input))

    while os.path.exists(tempName):
        nameHalves = (nameHalves[0] + '_', nameHalves[1])
        tempName = nameHalves[0] + nameHalves[1]

    try:
        if needsFix:
            zipLog = rebuildZip(input, tempName)
            deleteBadFrames(tempName, zipLog)

        # if 'zip -FF' gave us a 32bit zip file, copy it into a 64bit archive.
        fixed = tempName if needsFix else input
        if not isZip64(fixed):
            print('Copying from zip32 to zip64...')
            copy.main(fixed, tempName)

        if os.path.exists(tempName) and not os.path.samefile(tempName, output):
            os.rename(tempName, output)
    finally:
        if os.path.exists(tempName) and not os.path.samefile(tempName, output):
            os.remove(tempName)

if __name__ == '__main__': main(**vars(parser.parse_args()))



# distutils: language = c++

from libcpp.string cimport string
from libcpp.vector cimport vector
from cython.operator cimport dereference as deref

cimport numpy
from libcpp cimport bool

cdef extern from "numpy/arrayobject.h":
    cdef int PyArray_SetBaseObject(numpy.ndarray arr, obj)

cdef extern from "../src/SharedArray.hpp" namespace "gtar_pymodule::gtar":
    cdef cppclass SharedArray[T]:
        # ctypedef (T*) iterator

        SharedArray()
        SharedArray(T*, size_t)
        SharedArray(const SharedArray[T]&)
        # ~SharedArray()

        void copy(const SharedArray[T]&)
        bool isNull()
        T* begin()
        T* end()

        T *get()
        size_t size() const
        void release()
        T *disown()

        void swap(SharedArray[T]&)
        T &operator[]
        const T &operator[] const

cdef extern from "../src/Archive.hpp" namespace "gtar_pymodule::gtar":
    cdef enum OpenMode:
        Read
        Write
        Append

    cdef enum CompressMode:
        NoCompress
        FastCompress
        MediumCompress
        SlowCompress

cdef extern from "../src/Record.hpp" namespace "gtar_pymodule::gtar":
    cdef enum Behavior:
        Constant
        Discrete
        Continuous

    cdef enum Format:
        Float32
        Float64
        Int32
        Int64
        UInt8
        UInt32
        UInt64

    cdef enum Resolution:
        Text
        Uniform
        Individual

    cdef cppclass Record:
        Record()
        Record(const string&)
        Record(const string&, const string&, const string&,
               Behavior, Format, Resolution)
        Record(const Record&)

        void copy(const Record&)

        string nullifyIndex()
        Record withNullifiedIndex() const

        string getPath() const

        Behavior getBehavior() const
        string getGroup() const
        string getName() const
        string getIndex() const
        Format getFormat() const
        Resolution getResolution() const
        void setIndex(const string&)

cdef extern from "../src/GTAR.hpp" namespace "gtar_pymodule::gtar":
    cdef cppclass GTAR:
        cppclass BulkWriter:
            BulkWriter(GTAR&)

            void writeString(const string&, const string&, CompressMode) except +
            void writeBytes(const string&, const vector[char]&, CompressMode) except +
            void writePtr(const string&, const void*, const size_t, CompressMode) except +

        GTAR(const string&, OpenMode) except +

        void close()
        void writeString(const string&, const string&, CompressMode) except +
        void writeBytes(const string&, const vector[char]&, CompressMode) except +
        void writePtr(const string&, const void*, const size_t, CompressMode) except +

        # SharedArray[T] readIndividual[T](const string&)
        SharedArray[char] readBytes(const string&) except +

        vector[Record] getRecordTypes() const
        vector[string] queryFrames(const Record&) const

cdef extern from "../src/ZipArchive.hpp" namespace "gtar_pymodule::gtar":
     bool isZip64(const string&) except +



import sys
import gtar
import numpy as np

def test_widths():
    with gtar.GTAR('test.zip', 'w') as arch:
        arch.writePath('position.f32.ind', np.random.rand(6))

    with gtar.GTAR('test.zip', 'r') as arch:
        posx3 = arch.readPath('position.f32.ind')

        gtar.widths['position'] = 2
        posx2 = arch.readPath('position.f32.ind')

        gtar.widths.clear()
        posFlat = arch.readPath('position.f32.ind')

    success = posx3.shape == (2, 3) and posx2.shape == (3, 2) and posFlat.shape == (6,)

    assert success
    return success

def main():
    success = test_widths()

    sys.exit(not success)

if __name__ == '__main__': main()



"""
Copy each record from one getar-formatted file to another.

::

   usage: python -m gtar.copy [-h] input output

   Command-line archive copier or translator

   positional arguments:
     input           Input file to read
     output          File to write to

   optional arguments:
     -h, --help      show this help message and exit
"""

import argparse

from . import cat

parser = argparse.ArgumentParser(
    description='Command-line archive copier or translator')
parser.add_argument('input',
                    help='Input file to read')
parser.add_argument('output',
                    help='File to write to')

def main(input, output):
    """Copy each record from one getar-formatted file to another.

    :param input: Input filename
    :param output: Output filename (can be the same as input)

    """
    cat.main(inputs=[input], output=output)

if __name__ == '__main__': main(**vars(parser.parse_args()))



#!/usr/bin/env python

import os, subprocess, sys
from distutils.command.build_ext import build_ext
from distutils.sysconfig import get_config_var
from distutils.version import LooseVersion
from setuptools import Extension, setup
import numpy
import platform

long_description = """
libgetar is a library to read and write GEneric Trajectory ARchives, a
binary data format designed for efficient, extensible storage of
trajectory data.

For more details, consult the documentation at https://libgetar.readthedocs.io .
"""

# From https://github.com/pandas-dev/pandas/pull/24274:
# For mac, ensure extensions are built for macos 10.9 when compiling on a
# 10.9 system or above, overriding distuitls behaviour which is to target
# the version that python was built for. This may be overridden by setting
# MACOSX_DEPLOYMENT_TARGET before calling setup.py
if sys.platform == 'darwin':
    if 'MACOSX_DEPLOYMENT_TARGET' not in os.environ:
        current_system = LooseVersion(platform.mac_ver()[0])
        python_target = LooseVersion(get_config_var('MACOSX_DEPLOYMENT_TARGET'))
        if python_target < '10.9' and current_system >= '10.9':
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = '10.9'

with open('gtar/version.py') as version_file:
    exec(version_file.read())

macros = []
extra_args = []
include_dirs = [numpy.get_include(), 'lz4', 'miniz', 'sqlite3']
sources = [
    'src/Archive.cpp',
    'src/DirArchive.cpp',
    'src/GTAR.cpp',
    'src/Record.cpp',
    'src/SqliteArchive.cpp',
    'src/TarArchive.cpp',
    'src/ZipArchive.cpp',
    'lz4/lz4.c',
    'lz4/lz4hc.c',
    'miniz/miniz.cpp',
    'sqlite3/sqlite3.c',
]

# prefix module to avoid collisions with builtin versions of libgetar
macros.append(('GTAR_NAMESPACE_PARENT', 'gtar_pymodule'))

# set max length manually
macros.append(('SQLITE_MAX_LENGTH', '1000000000'))

if '--disable-read-check' in sys.argv:
    macros.append(('MINIZ_DISABLE_ZIP_READER_CRC32_CHECKS', None))
    sys.argv.remove('--disable-read-check')

if '--debug' in sys.argv:
    extra_args.extend(['-O0', '-g'])
    sys.argv.remove('--debug')

if '--cython' in sys.argv:
    from Cython.Build import cythonize
    sys.argv.remove('--cython')

    def myCythonize(macros, *args, **kwargs):
        result = cythonize(*args, **kwargs)
        for r in result:
            r.define_macros.extend(macros)
            r.include_dirs.extend(include_dirs)
            r.extra_compile_args.extend(extra_args)
            r.extra_link_args.extend(extra_args)
            r.sources.extend(sources)

        return result

    modules = myCythonize(macros, 'gtar/_gtar.pyx')
else:
    sources.append('gtar/_gtar.cpp')
    modules = [Extension('gtar._gtar', sources=sources,
                         define_macros=macros, extra_compile_args=extra_args,
                         extra_link_args=extra_args, include_dirs=include_dirs)]

setup(name='gtar',
      version=__version__,
      description='Read and write archive-based trajectory formats',
      author='Matthew Spellings',
      author_email='mspells@umich.edu',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Topic :: System :: Archiving',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Chemistry',
          'Topic :: Scientific/Engineering :: Physics',
      ],
      ext_modules=modules,
      license='MIT',
      long_description=long_description,
      packages=['gtar'],
      project_urls={
          'Documentation': 'http://libgetar.readthedocs.io/',
          'Source': 'https://github.com/glotzerlab/libgetar'
          },
      install_requires=['numpy'],
      url='http://libgetar.readthedocs.io/',
)



import functools
import unittest
import sys
import gtar
import numpy as np

SUFFIXES = ['.tar', '.zip', '.sqlite', '/']

SPECIAL_SUFFIX_NAMES = {'/': 'directory'}

def suffixed(f, suffix):
    """Decorator to instantiate a class method with a suffix"""

    def result(self):
        return f(self, suffix)

    result.__doc__ = f.__doc__
    return result

class MultiSuffixMeta(type):
    """Metaclass to instantiate tests for multiple file suffixes"""
    def __new__(cls, name, bases, dct):
        to_specialize = [key for key in dct if key.startswith('test_')]

        for fun_name in to_specialize:
            original_function = dct[fun_name]
            for suffix in SUFFIXES:
                name_suffix = SPECIAL_SUFFIX_NAMES.get(suffix, suffix[1:])
                new_name = '{}_{}'.format(fun_name, name_suffix)
                dct[new_name] = suffixed(original_function, suffix)
            del dct[fun_name]

        return super(MultiSuffixMeta, cls).__new__(cls, name, bases, dct)

class TestGTAR(unittest.TestCase):
    def test_group_selection(self, suffix):
        if suffix == '/':
            raise unittest.SkipTest(
                'directory support is currently experimental'
                ' and more robust group parsing needs to be added')

        with gtar.GTAR('test' + suffix, 'w') as arch:
            arch.writeStr('prefix/test.txt', 'sample text')

        with gtar.GTAR('test' + suffix, 'r') as arch:
            self.assertEqual(1, len(arch.framesWithRecordsNamed(
                'test.txt')[1]))
            self.assertEqual(1, len(arch.framesWithRecordsNamed(
                'test.txt', group='prefix')[1]))
            self.assertEqual(1, len(arch.framesWithRecordsNamed(
                'test.txt', group_prefix='p')[1]))

            self.assertEqual(0, len(arch.framesWithRecordsNamed(
                'test.txt', group_prefix='t')[1]))
            self.assertEqual(0, len(arch.framesWithRecordsNamed(
                'test.txt', group='prefi')[1]))

            self.assertEqual(1, len(list(arch.recordsNamed(
                'test.txt'))))
            self.assertEqual(1, len(list(arch.recordsNamed(
                'test.txt', group='prefix'))))
            self.assertEqual(1, len(list(arch.recordsNamed(
                'test.txt', group_prefix='p'))))

            self.assertEqual(0, len(list(arch.recordsNamed(
                'test.txt', group_prefix='t'))))
            self.assertEqual(0, len(list(arch.recordsNamed(
                'test.txt', group='prefi'))))

            self.assertEqual(
                gtar.Record('prefix/test.txt'), arch.staticRecordNamed(
                    'test.txt', group='prefix'))
            self.assertEqual(
                gtar.Record('prefix/test.txt'), arch.staticRecordNamed(
                    'test.txt', group_prefix='p'))

    def test_readThenWrite(self, suffix):
        records = {'test.txt': 'test string foo\n',
                   'blah.txt': 'another string',
                   'frames/0/otherblah.f32.ind': np.random.rand(5),
                   'frames/0/anotherblah.f64.ind': np.random.rand(7),
                   'frames/0/number.i32.uni': 14}

        convertedRecords = {'frames/0/otherblah.f32.ind':
                            np.array(records['frames/0/otherblah.f32.ind'],
                                     dtype=np.float32)}

        with gtar.GTAR('test' + suffix, 'w') as arch:
            for path in records:
                arch.writePath(path, records[path])

        for mode in ['r', 'a']:
            with gtar.GTAR('test' + suffix, mode) as arch:
                for path in records:
                    self.assertTrue(np.all(
                        arch.readPath(path) == convertedRecords.get(path, records[path])))

    def test_readAndWrite(self, suffix):
        records = {'test.txt': 'test string foo\n',
                   'blah.txt': 'another string',
                   'frames/0/otherblah.f32.ind': np.random.rand(5),
                   'frames/0/anotherblah.f64.ind': np.random.rand(7),
                   'frames/0/number.i32.uni': 14}

        convertedRecords = {'frames/0/otherblah.f32.ind':
                            np.array(records['frames/0/otherblah.f32.ind'],
                                     dtype=np.float32)}

        written = []
        with gtar.GTAR('test' + suffix, 'w') as arch:
            for path in records:
                arch.writePath(path, records[path])
                written.append(path)

                for readpath in sorted(written):
                    self.assertTrue(np.all(
                        arch.readPath(readpath) == convertedRecords.get(readpath, records[readpath])))

    def test_readAndWriteRecords(self, suffix):
        with gtar.GTAR('test' + suffix, 'w') as arch:
            reclen = len(arch.getRecordTypes())
            for idx in range(10):
                arch.writeStr(str(idx), '{}_contents'.format(idx))

                self.assertEqual(reclen + 1, len(arch.getRecordTypes()))
                reclen = len(arch.getRecordTypes())

    def test_overwrite(self, suffix):
        with gtar.GTAR('test' + suffix, 'w') as arch:
            arch.writeStr('test.txt', 'bad')
            arch.writeStr('test.txt', 'good')

            self.assertEqual(arch.readStr('test.txt'), 'good')

        with gtar.GTAR('test' + suffix, 'r') as arch:
            self.assertEqual(arch.readStr('test.txt'), 'good')

    def test_read_closed(self, suffix):
        with gtar.GTAR('test' + suffix, 'w') as arch:
            arch.writeStr('test.txt', 'good')

        # throw when reading from a closed archive
        with self.assertRaises(RuntimeError):
            arch.readStr('test.txt')

        # readPath goes through getRecord
        with self.assertRaises(RuntimeError):
            arch.readPath('test.txt')

    def test_write_readonly(self, suffix):
        with gtar.GTAR('test' + suffix, 'w') as arch:
            arch.writeStr('test.txt', 'good')

        with gtar.GTAR('test' + suffix, 'r') as arch:
            with self.assertRaises(RuntimeError):
                arch.writeStr('test.txt', 'bad')

TestGTAR = MultiSuffixMeta(
    TestGTAR.__name__, TestGTAR.__bases__, dict(TestGTAR.__dict__))

if __name__ == '__main__':
    unittest.main()
